# Message control annotations.
#
# Author::    Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>
# Copyright:: Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
# License::   GPLv3+: GNU General Public License version 3 or later
#
# Owner::     Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>

#--
#     ___    ____  __    ___   _________
#    /   |  / _  |/ /   / / | / /__  __/           Source Code Static Analyzer
#   / /| | / / / / /   / /  |/ /  / /                   AdLint - Advanced Lint
#  / __  |/ /_/ / /___/ / /|  /  / /
# /_/  |_|_____/_____/_/_/ |_/  /_/   Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
#
# This file is part of adlint-postfilter.
#
# adlint-postfilter is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# adlint-postfilter is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# adlint-postfilter.  If not, see <http://www.gnu.org/licenses/>.
#
#++

require "adlint/postfilter/suppress"

module AdLint #:nodoc:
module Postfilter #:nodoc:

  class MessageControlAnnotation
    def initialize(message_controller_str)
      @message_controller = MessageController.new(message_controller_str)
    end

    attr_reader :message_controller
  end

  class TranslationUnitWideAnnotation < MessageControlAnnotation
    def self.detect(str, composing_fpaths)
      str =~ /ADLINT:TUNIT:\[.*\]/ ? new($&, composing_fpaths) : nil
    end

    def initialize(message_controller_str, composing_fpaths)
      super(message_controller_str)
      @composing_fpaths = composing_fpaths
    end
    private_class_method :new

    def generate_suppressions
      message_controller.message_control_list.map { |mesg_ctrl|
        @composing_fpaths.map do |fpath|
          FileWiseSuppression.new(mesg_ctrl.message_id, fpath,
                                  mesg_ctrl.suppression_activation?)
        end
      }.flatten
    end
  end

  class LineWiseAnnotation < MessageControlAnnotation
    def self.detect(str, fpath, line_no)
      str =~ /(?:ADLINT:LINE:|ADLINT::)\[.*\]/ ? new($&, fpath, line_no) : nil
    end

    def initialize(message_controller_str, fpath, line_no)
      super(message_controller_str)
      @fpath, @line_no = fpath, line_no
    end
    private_class_method :new

    def generate_suppressions
      message_controller.message_control_list.map do |mesg_ctrl|
        LineWiseSuppression.new(mesg_ctrl.message_id, @fpath, @line_no,
                                mesg_ctrl.suppression_activation?)
      end
    end
  end

  class MessageController
    def initialize(message_controller_str)
      @message_control_list = parse_message_controller(message_controller_str)
    end

    attr_reader :message_control_list

    private
    def parse_message_controller(message_controller_str)
      if message_controller_str =~ /\[(.*)\]/
        parse_message_control_list($1)
      else
        []
      end
    end

    def parse_message_control_list(message_control_list_str)
      message_control_list_str.scan(/([+-])([WC]\d{4})/).map do |op, mesg_id|
        case op
        when "-"
          MessageSuppressionActivation.new(mesg_id)
        when "+"
          MessageSuppressionDeactivation.new(mesg_id)
        end
      end
    end
  end

  class MessageControl
    def initialize(message_id_str)
      @message_id = message_id_str.to_sym
    end

    attr_reader :message_id

    def suppression_activation?
      subclass_responsibility
    end

    def suppression_deactivation?
      !suppression_activation?
    end
  end

  class MessageSuppressionActivation < MessageControl
    def suppression_activation?
      true
    end
  end

  class MessageSuppressionDeactivation < MessageControl
    def suppression_activation?
      false
    end
  end

end
end
