# AdLint analysis commands.
#
# Author::    Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>
# Copyright:: Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
# License::   GPLv3+: GNU General Public License version 3 or later
#
# Owner::     Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>

#--
#     ___    ____  __    ___   _________
#    /   |  / _  |/ /   / / | / /__  __/           Source Code Static Analyzer
#   / /| | / / / / /   / /  |/ /  / /                   AdLint - Advanced Lint
#  / __  |/ /_/ / /___/ / /|  /  / /
# /_/  |_|_____/_____/_/_/ |_/  /_/   Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
#
# This file is part of adlint-postfilter.
#
# adlint-postfilter is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# adlint-postfilter is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# adlint-postfilter.  If not, see <http://www.gnu.org/licenses/>.
#
#++

require "fileutils"
require "getoptlong"
require "pathname"

require "adlint/postfilter/config"
require "adlint/postfilter/content"
require "adlint/postfilter/filter"

module AdLint #:nodoc:
module Postfilter #:nodoc:

  class AnalysisCommand
    def self.for(command_line, config_fpath, src_vpath = nil)
      case name = command_line.split.first
      when "adlint"
        ADLINT.new(command_line, config_fpath, src_vpath)
      when "adlint_sma"
        ADLINT_SMA.new(command_line, config_fpath, src_vpath)
      when "adlint_cma"
        ADLINT_CMA.new(command_line, config_fpath, src_vpath)
      when "adlint_chk"
        ADLINT_CHK.new(command_line, config_fpath, src_vpath)
      else
        raise "unknown command `#{name}'"
      end
    end

    def initialize(command_line, config_fpath, src_vpath = nil)
      @command_line = command_line
      @config_fpath = config_fpath
      @src_vpath = src_vpath

      @traits_fpath, @strip_num, @output_dpath, @src_fpaths =
        parse_adlint_command_line(@command_line, @src_vpath)
    end

    def execute
      status, stderr_content = exec_adlint_command(@command_line)
      filter_results(create_filters, stderr_content)
      status
    end

    private
    def exec_adlint_command(command_line)
      system("#{command_line} 2>#{stderr_fpath}")
      return $?, StderrContent.new(File.read(stderr_fpath))
    ensure
      FileUtils.rm_f(stderr_fpath)
    end

    def filter_results(filters, stderr_content)
      filters.each { |filter| filter.execute(stderr_content) }
      stderr_content.overwrite!
    end

    def create_filters
      subclass_responsibility
    end

    def parse_adlint_command_line(command_line, src_vpath = nil)
      parser = GetoptLong.new(
        ["--traits", "-t", GetoptLong::REQUIRED_ARGUMENT],
        ["--list-file", "-l", GetoptLong::REQUIRED_ARGUMENT],
        ["--output-dir", "-o", GetoptLong::REQUIRED_ARGUMENT],
        ["--strip", "-p", GetoptLong::REQUIRED_ARGUMENT],
        ["--verbose", "-v", GetoptLong::NO_ARGUMENT])
      parser.quiet = true

      begin
        traits_fpath = nil
        output_dpath = nil
        strip_num = src_vpath ? src_vpath.components.size : 0

        ARGV.replace(command_line.split).shift

        parser.each_option do |optname, optarg|
          case optname
          when "--traits"
            traits_fpath = Pathname.new(optarg)
          when "--output-dir"
            output_dpath = Pathname.new(optarg)
          when "--strip"
            strip_num = optarg.to_i
          end
        end
      rescue
        # NOTE: If the AdLint's command line is invalid, child analysis command
        #       will encounter the same problem and will notice it.
        #       So, nothing to be done.
      end

      src_fpaths = ARGV.map { |str|
        if str.end_with?(".c.met.csv")
          @src_vpath.join(Pathname.new(str.sub(/.met.csv\z/, "")))
        else
          Pathname.new(str)
        end
      }

      return traits_fpath, strip_num, output_dpath, src_fpaths
    end

    def create_sma_filters
      create_sma_configs.map { |sma_config| MessageFilter.new(sma_config) }
    end

    def create_cma_filter
      MessageFilter.new(create_cma_config)
    end

    def create_sma_configs
      @src_fpaths.map do |src_fpath|
        SmaConfig.new(@config_fpath, @traits_fpath, src_fpath, @strip_num,
                      @output_dpath)
      end
    end

    def create_cma_config
      CmaConfig.new(@config_fpath, @traits_fpath, @strip_num, @output_dpath,
                    create_sma_configs)
    end

    def stderr_fpath
      Pathname.new("./adlint_postfilter.#{$$}.stderr")
    end
  end

  class ADLINT < AnalysisCommand
    private
    def create_filters
      create_sma_filters + [create_cma_filter]
    end
  end

  class ADLINT_SMA < AnalysisCommand
    private
    def create_filters
      create_sma_filters
    end
  end

  class ADLINT_CMA < AnalysisCommand
    private
    def create_filters
      [create_cma_filter]
    end
  end

  class ADLINT_CHK < ADLINT_SMA; end

end
end
