# Filtering contents.
#
# Author::    Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>
# Copyright:: Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
# License::   GPLv3+: GNU General Public License version 3 or later
#
# Owner::     Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>

#--
#     ___    ____  __    ___   _________
#    /   |  / _  |/ /   / / | / /__  __/           Source Code Static Analyzer
#   / /| | / / / / /   / /  |/ /  / /                   AdLint - Advanced Lint
#  / __  |/ /_/ / /___/ / /|  /  / /
# /_/  |_|_____/_____/_/_/ |_/  /_/   Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
#
# This file is part of adlint-postfilter.
#
# adlint-postfilter is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# adlint-postfilter is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# adlint-postfilter.  If not, see <http://www.gnu.org/licenses/>.
#
#++

require "csv"
require "pathname"

module AdLint #:nodoc:
module Postfilter #:nodoc:

  class FilteringContent < Array
    def overwrite!
      subclass_responsibility
    end
  end

  class MessageFileContent < FilteringContent
    def initialize(fpath)
      @fpath = fpath
      lines = File.read(fpath).lines.each_with_object([]) { |line, ary|
        if line[0] == "V"
          ary.push(OrdinaryLine.new(line.chomp))
        else
          ary.push(CsvMessageLine.new(line.chomp))
        end
      }
      super(lines)
    end

    def overwrite!
      File.open(@fpath, "w") { |io| io.puts self.compact }
    end
  end

  class StderrContent < FilteringContent
    def initialize(str)
      lines = str.lines.each_with_object([]) { |line, ary|
        if line =~ /:(?:warning|context):/
          ary.push(PlainMessageLine.new(line.chomp))
        else
          ary.push(OrdinaryLine.new(line.chomp))
        end
      }
      super(lines)
    end

    def overwrite!
      $stderr.puts self.compact
    end
  end

  class Line < String
    def filterable?
      subclass_responsibility
    end
  end

  class OrdinaryLine < Line
    def filterable?
      false
    end
  end

  class MessageLine < Line
    def initialize(str, fpath_str, line_no_str, message_id_str)
      super(str)

      @fpath = Pathname.new(fpath_str).realpath
      @line_no = line_no_str ? line_no_str.to_i : nil
      @message_id = message_id_str.to_sym
    end

    attr_reader :fpath
    attr_reader :line_no
    attr_reader :message_id

    def filterable?
      true
    end

    def match_with?(suppression)
      suppression._match_with?(self)
    end
  end

  class CsvMessageLine < MessageLine
    def initialize(str)
      fields = CSV.parse(str).first
      super(str, fields.fetch(1), fields.fetch(2), fields.fetch(4))
    end
  end

  class PlainMessageLine < MessageLine
    def initialize(str)
      super(str, *str.scan(/\A(.*):(?:(\d*):\d*|):.*:(.*):/).first)
    end
  end

end
end
