# Message suppression.
#
# Author::    Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>
# Copyright:: Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
# License::   GPLv3+: GNU General Public License version 3 or later
#
# Owner::     Yutaka Yanoh <mailto:yanoh@users.sourceforge.net>

#--
#     ___    ____  __    ___   _________
#    /   |  / _  |/ /   / / | / /__  __/           Source Code Static Analyzer
#   / /| | / / / / /   / /  |/ /  / /                   AdLint - Advanced Lint
#  / __  |/ /_/ / /___/ / /|  /  / /
# /_/  |_|_____/_____/_/_/ |_/  /_/   Copyright (C) 2010-2012, OGIS-RI Co.,Ltd.
#
# This file is part of adlint-postfilter.
#
# adlint-postfilter is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# adlint-postfilter is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# adlint-postfilter.  If not, see <http://www.gnu.org/licenses/>.
#
#++

module AdLint #:nodoc:
module Postfilter #:nodoc:

  class MessageSuppression
    def _match_with?(message_line)
      subclass_responsibility
    end
  end

  class AdLintInitialHeaderSuppression < MessageSuppression
    def initialize(cinit_fpath, pinit_fpath)
      @cinit_fpath = cinit_fpath
      @pinit_fpath = pinit_fpath
    end

    def _match_with?(message_line)
      message_line.fpath == @cinit_fpath || message_line.fpath == @pinit_fpath
    end
  end

  class PlatformHeaderSuppression < MessageSuppression
    def initialize(project_root_dpath)
      @project_root_dpath = project_root_dpath.realpath
    end

    def _match_with?(message_line)
      !message_line.fpath.to_s.start_with?(@project_root_dpath.to_s)
    end
  end

  class MessageWiseSuppression < MessageSuppression
    def initialize(message_id)
      @message_id = message_id
    end

    attr_reader :message_id

    def _match_with?(message_line)
      message_line.message_id == @message_id
    end
  end

  class ProjectWideSuppression < MessageWiseSuppression; end

  module Deactivatable
    attr_writer :active

    def active?
      @active
    end
  end

  class FileWiseSuppression < MessageWiseSuppression
    include Deactivatable

    def initialize(message_id, fpath, active)
      super(message_id)
      @fpath = fpath
      self.active = active
    end

    attr_reader :fpath

    def _match_with?(message_line)
      if self.active?
        super && message_line.fpath == @fpath
      else
        false
      end
    end
  end

  class LineWiseSuppression < MessageWiseSuppression
    include Deactivatable

    def initialize(message_id, fpath, line_no, active)
      super(message_id)
      @fpath, @line_no = fpath, line_no
      self.active = active
    end

    attr_reader :fpath
    attr_reader :line_no

    def _match_with?(message_line)
      if self.active?
        super &&
          message_line.fpath == @fpath && message_line.line_no == @line_no
      else
        false
      end
    end
  end

  class MessageSuppressionList
    def initialize(initial_header_suppression, platform_header_suppression,
                   project_wide_suppressions,
                   tunit_wide_suppressions, line_wise_suppressions)
      @initial_header_suppression = initial_header_suppression
      @platform_header_suppression = platform_header_suppression
      @p_index = create_p_index(project_wide_suppressions)
      @t_index = create_t_index(tunit_wide_suppressions)
      @l_index = create_l_index(line_wise_suppressions)
    end

    def apply_to(content)
      content.each_with_index do |line, index|
        if line.filterable?
          supps = suppressions_for(line.message_id, line.fpath, line.line_no)
          if supps.any? { |supp| line.match_with?(supp) }
            content[index] = nil
          end
        end
      end
    end

    private
    def suppressions_for(message_id, fpath, line_no)
      p_hash = select_project_wide_suppressions(message_id)
      t_hash = select_tunit_wide_suppressions(message_id, fpath)
      l_hash = select_line_wise_suppressions(message_id, fpath, line_no)
      [@initial_header_suppression, @platform_header_suppression] +
        p_hash.merge(t_hash).merge(l_hash).values
    end

    def select_project_wide_suppressions(message_id)
      (supp = @p_index[message_id]) ? {message_id => supp} : {}
    end

    def select_tunit_wide_suppressions(message_id, fpath)
      (supp = @t_index[message_id][fpath]) ? {message_id => supp} : {}
    end

    def select_line_wise_suppressions(message_id, fpath, line_no)
      @l_index[message_id][fpath].each_with_object({}) do |supp, result|
        if supp.line_no == line_no
          result[supp.message_id] = supp
        end
      end
    end

    def create_p_index(suppressions)
      suppressions.each_with_object({}) do |supp, index|
        index[supp.message_id] = supp
      end
    end

    def create_t_index(suppressions)
      index = Hash.new { |mesg_id_hash, mesg_id| mesg_id_hash[mesg_id] = {} }
      suppressions.each do |supp|
        index[supp.message_id][supp.fpath] = supp
      end
      index
    end

    def create_l_index(suppressions)
      index = Hash.new { |mesg_id_hash, mesg_id|
        mesg_id_hash[mesg_id] = Hash.new { |fpath_hash, fpath|
          fpath_hash[fpath] = []
        }
      }
      suppressions.each do |supp|
        index[supp.message_id][supp.fpath].push(supp)
      end
      index
    end
  end

end
end
